# -*- coding: utf-8 -*-
# snapshottest: v1 - https://goo.gl/zC4yUc
from __future__ import unicode_literals

from snapshottest import Snapshot


snapshots = Snapshot()

snapshots['test_help 1'] = '''usage:  comment [-h] COMMENT [COMMENT ...]

positional arguments:
  COMMENT     Comment

optional arguments:
  -h, --help  show this help message and exit
'''

snapshots['test_help 2'] = '''usage:  steps [-h]

optional arguments:
  -h, --help  show this help message and exit
'''

snapshots['test_help 3'] = '''usage:  env [-h] --var KEY VALUE

optional arguments:
  -h, --help       show this help message and exit
  --var KEY VALUE  A map of environment variables for this pipeline.
'''

snapshots['test_help 4'] = '''usage:  command [-h] --command COMMAND [COMMAND ...] [--label LABEL]
                [--branches BRANCH_PATTERN [BRANCH_PATTERN ...]]
                [--env KEY VALUE] [--agents KEY VALUE]
                [--artifact-paths GLOB_OR_PATH [GLOB_OR_PATH ...]]
                [--parallelism POSITIVE_NUMBER] [--concurrency POSITIVE_INT]
                [--concurrency-group GROUP_NAME]
                [--timeout-in-minutes TIMEOUT] [--skip BOOL_OR_STRING]
                [--retry {automatic,manual}]
                [--retry-automatic-exit-status INT_OR_STAR]
                [--retry-automatic-limit POSITIVE_INT]
                [--retry-automatic-tuple INT_OR_STAR POSITIVE_INT]
                [--retry-manual-allowed] [--no-retry-manual-allowed]
                [--retry-manual-reason REASON]
                [--retry-manual-permit-on-passed]
                [--no-retry-manual-permit-on-passed]
                [--plugin PLUGIN [KEY_VALUE_PAIR ...]]

optional arguments:
  -h, --help            show this help message and exit
  --command COMMAND [COMMAND ...]
                        The shell command/s to run during this step.
  --label LABEL         The label that will be displayed in the pipeline
                        visualisation in Buildkite. Supports emoji.
  --branches BRANCH_PATTERN [BRANCH_PATTERN ...]
                        The branch pattern defining which branches will
                        include this step in their builds.
  --env KEY VALUE       A map of environment variables for this step.
  --agents KEY VALUE    A map of meta-data keys to values to target specific
                        agents for this step.
  --artifact-paths GLOB_OR_PATH [GLOB_OR_PATH ...]
                        The glob path or paths where artifacts from this step
                        will be uploaded.
  --parallelism POSITIVE_NUMBER
                        The number of parallel jobs that will be created based
                        on this step.
  --concurrency POSITIVE_INT
                        The maximum number of jobs created from this step that
                        are allowed to run at the same time. Requires
                        --concurrency-group.
  --concurrency-group GROUP_NAME
                        A unique name for the concurrency group that you are
                        creating with the concurrency attribute.
  --timeout-in-minutes TIMEOUT
                        The number of minutes a job created from this step is
                        allowed to run. If the job does not finish within this
                        limit, it will be automatically cancelled and the
                        build will fail.
  --skip BOOL_OR_STRING
                        Whether to skip this step or not.
  --retry {automatic,manual}
                        The conditions for retrying this step.
  --retry-automatic-exit-status INT_OR_STAR
                        The exit status number that will cause this job to
                        retry.
  --retry-automatic-limit POSITIVE_INT
                        The number of times this job can be retried. The
                        maximum value this can be set to is 10.
  --retry-automatic-tuple INT_OR_STAR POSITIVE_INT
                        The exit status number that will cause this job to
                        retry and a limit to go with it.
  --retry-manual-allowed
                        This job can be retried manually.
  --no-retry-manual-allowed
                        This job can not be retried manually.
  --retry-manual-reason REASON
                        A string that will be displayed in a tooltip on the
                        Retry button in Buildkite.
  --retry-manual-permit-on-passed
                        This job can be retried after it has passed.
  --no-retry-manual-permit-on-passed
                        This job can not be retried after it has passed.
  --plugin PLUGIN [KEY_VALUE_PAIR ...]
                        A plugin to run with this step. Optionally key/value
                        pairs for the plugin.
'''

snapshots['test_steps 1'] = '''steps:
'''

snapshots['test_main 1'] = '''  - command: x
'''

snapshots['test_retry_automatic 1'] = '''  - command: cmd
    retry:
      automatic: true
'''

snapshots['test_retry_automatic_limit 1'] = '''  - command: cmd
    retry:
      automatic:
        limit: 2
'''

snapshots['test_retry_automatic_limit_11 1'] = '''  - command: cmd
    retry:
      automatic:
        limit: 10
'''

snapshots['test_retry_automatic_exit_status_star 1'] = '''  - command: cmd
    retry:
      automatic:
        exit_status: '*'
'''

snapshots['test_retry_automatic_exit_status_number 1'] = '''  - command: cmd
    retry:
      automatic:
        exit_status: 1
'''

snapshots['test_retry_automatic_exit_status_and_limit 1'] = '''  - command: cmd
    retry:
      automatic:
        exit_status: 1
        limit: 2
'''

snapshots['test_retry_automatic_tuple_0 1'] = '''  - command: cmd
    retry:
      automatic: true
'''

snapshots['test_retry_automatic_tuple_n 1'] = '''  - command: cmd
    retry:
      automatic:
        - exit_status: '*'
          limit: 2
        - exit_status: 1
          limit: 3
'''

snapshots['test_retry_manual 1'] = '''  - command: cmd
    retry:
      manual: true
'''

snapshots['test_retry_manual_allowed 1'] = '''  - command: cmd
    retry:
      manual: true
'''

snapshots['test_retry_manual_allowed 2'] = '''  - command: cmd
    retry:
      manual:
        allowed: false
'''

snapshots['test_retry_reason 1'] = '''  - command: cmd
    retry:
      manual:
        reason: Some reason why
'''

snapshots['test_retry_manual_permit_on_passed 1'] = '''  - command: cmd
    retry:
      manual:
        permit_on_passed: true
'''

snapshots['test_retry_manual_permit_on_passed 2'] = '''  - command: cmd
    retry:
      manual: true
'''

snapshots['test_command_1 1'] = '''  - command: my-command arg1 'arg 2'
'''

snapshots['test_command_n 1'] = '''  - command:
      - a
      - b
      - c
      - d
'''

snapshots['test_label 1'] = '''  - label: My label
    command: cmd
'''

snapshots['test_branches 1'] = '''  - command: cmd
    branches: master release-*
'''

snapshots['test_env 1'] = '''  - command: cmd
    env:
      a: b
      c: d
'''

snapshots['test_agents 1'] = '''  - command: cmd
    agents:
      npm: 'true'
      mvn: 'true'
'''

snapshots['test_artifact_paths_0 1'] = '''  - command: cmd
'''

snapshots['test_artifact_paths_1 1'] = '''  - command: cmd
    artifact_paths: logs/**/*;coverage/**/*
'''

snapshots['test_artifact_paths_n 1'] = '''  - command: cmd
    artifact_paths:
      - logs/**/*
      - coverage/**/*
'''

snapshots['test_parallelism 1'] = '''  - command: cmd
    parallelism: 4
'''

snapshots['test_parallelism_1 1'] = '''  - command: cmd
'''

snapshots['test_concurrency 1'] = '''  - command: cmd
    concurrency: 2
    concurrency_group: my/group
'''

snapshots['test_timeout_in_minutes_minus 1'] = '''  - command: cmd
'''

snapshots['test_timeout_in_minutes_0 1'] = '''  - command: cmd
'''

snapshots['test_timeout_in_minutes_1 1'] = '''  - command: cmd
    timeout_in_minutes: 1
'''

snapshots['test_skip_bool_true 1'] = '''  - command: cmd
    skip: true
'''

snapshots['test_skip_bool_false 1'] = '''  - command: cmd
'''

snapshots['test_skip_string 1'] = '''  - command: cmd
    skip: Some reason
'''

snapshots['test_env_all 1'] = '''env:
  a: b
  c: d
'''

snapshots['test_retry_manual_pop_manual_retry 1'] = '''  - command: cmd
    retry:
      manual:
        permit_on_passed: true

'''

snapshots['test_help 5'] = '''usage:  plugin [-h] [--name NAME] --plugin PLUGIN [KEY_VALUE_PAIR ...]

optional arguments:
  -h, --help            show this help message and exit
  --name NAME           Name of the plugin step
  --plugin PLUGIN [KEY_VALUE_PAIR ...]
                        A plugin to run. Optionally key/value pairs for the
                        plugin.
'''

snapshots['test_command_plugin_none 1'] = '''  - command: cmd
'''

snapshots['test_command_plugin_no_args 1'] = '''  - command: cmd
    plugins:
      org/repo#1.0.0:
'''

snapshots['test_command_plugin_1 1'] = '''  - command: cmd
    plugins:
      org/repo#1.0.0:
        a: b
        c: d
'''

snapshots['test_command_plugin_n 1'] = '''  - command: cmd
    plugins:
      org/repo#1.0.0:
        a: b
        c: d
      other_org/other_repo:
        e: f
        g: h=i
'''

snapshots['test_plugin_plugin_none 1'] = ''

snapshots['test_plugin_plugin_no_args 1'] = '''  - plugins:
      org/repo#1.0.0:
'''

snapshots['test_plugin_plugin_1 1'] = '''  - plugins:
      org/repo#1.0.0:
        a: b
        c: d
'''

snapshots['test_plugin_plugin_n 1'] = '''  - plugins:
      org/repo#1.0.0:
        a: b
        c: d
      other_org/other_repo:
        e: f
        g: h=i
'''

snapshots['test_plugin_name_attr 1'] = '''  - name: My plugin run
    plugins:
      org/repo#1.0.0:
'''

snapshots['test_help 6'] = '''usage:  wait [-h] [--continue-on-failure]

optional arguments:
  -h, --help            show this help message and exit
  --continue-on-failure
                        Continue even if previous steps failed
'''

snapshots['test_wait 1'] = '''  - wait
'''

snapshots['test_wait_continue_on_failure 1'] = '''  - wait:
    continue_on_failure: true
'''

snapshots['test_comment 1'] = '''# a
# b'''

snapshots['test_comment_multiline 1'] = '''# multiline
# comments
# are fun'''

snapshots['test_trigger_simple 1'] = '''  - trigger: my-pipeline
'''

snapshots['test_trigger_label 1'] = '''  - trigger: my-pipeline
    label: ':rocket: Deploy'
'''

snapshots['test_help 7'] = '''usage:  trigger [-h] [--label LABEL] [--async]
                [--branches BRANCH_PATTERN [BRANCH_PATTERN ...]]
                [--build-message MESSAGE] [--build-commit SHA]
                [--build-branch BRANCH] [--build-env KEY VALUE]
                [--build-meta-data KEY VALUE]
                PIPELINE

positional arguments:
  PIPELINE              Name of the pipeline to trigger

optional arguments:
  -h, --help            show this help message and exit
  --label LABEL         The label that will be displayed in the pipeline
                        visualisation in Buildkite. Supports emoji.
  --async               If given, the step will immediately continue,
                        regardless of the success of the triggered build.
  --branches BRANCH_PATTERN [BRANCH_PATTERN ...]
                        The branch pattern defining which branches will
                        include this step in their builds.
  --build-message MESSAGE
                        The message for the build. Supports emoji.
  --build-commit SHA    The commit hash for the build
  --build-branch BRANCH
                        The branch for the build
  --build-env KEY VALUE
                        A map of environment variables for the triggered
                        build.
  --build-meta-data KEY VALUE
                        A map of meta-data for the triggered build.
'''

snapshots['test_trigger_async 1'] = '''  - trigger: my-pipeline
    async: true
'''

snapshots['test_trigger_branches 1'] = '''  - trigger: my-pipeline
    branches: master release-*
'''

snapshots['test_trigger_build_branch 1'] = '''  - trigger: my-pipeline
    build:
      branch: master
'''

snapshots['test_trigger_build_message 1'] = '''  - trigger: my-pipeline
    build:
      message: Put the lime in the coconut
'''

snapshots['test_trigger_build_commit 1'] = '''  - trigger: my-pipeline
    build:
      commit: c0ffee
'''

snapshots['test_trigger_build_env 1'] = '''  - trigger: my-pipeline
    build:
      env:
        a: b
'''

snapshots['test_trigger_build_meta_data 1'] = '''  - trigger: my-pipeline
    build:
      meta_data:
        a: b
'''

snapshots['test_empty_command 1'] = '''usage:  [-h] [--version] [-v] [-vv]
        {comment,steps,env,command,plugin,wait,trigger,block} ...

Generate pipeline YAML for Buildkite

optional arguments:
  -h, --help            show this help message and exit
  --version             show program's version number and exit
  -v, --verbose         set loglevel to INFO
  -vv, --very-verbose   set loglevel to DEBUG

subcommands:
  valid subcommands

  {comment,steps,env,command,plugin,wait,trigger,block}
                        additional help
'''

snapshots['test_help 8'] = '''usage:  block [-h] [--prompt PROMPT]
              [--branches BRANCH_PATTERN [BRANCH_PATTERN ...]]
              [--field-text KEY LABEL HINT REQUIRED DEFAULT]
              [--field-select KEY LABEL HINT REQUIRED DEFAULT KEY_VALUE_PAIRS [KEY_VALUE_PAIRS ...]
              [KEY LABEL HINT REQUIRED DEFAULT KEY_VALUE_PAIRS [KEY_VALUE_PAIRS ...]
              ...]]
              LABEL

positional arguments:
  LABEL                 Label of the block step. Supports emoji.

optional arguments:
  -h, --help            show this help message and exit
  --prompt PROMPT       The instructional message displayed in the dialog box
                        when the unblock step is activated.
  --branches BRANCH_PATTERN [BRANCH_PATTERN ...]
                        The branch pattern defining which branches will
                        include this step in their builds.
  --field-text KEY LABEL HINT REQUIRED DEFAULT
                        A text field. KEY is the meta-data key that stores the
                        field's input. LABEL is the label of the field. HINT
                        is the explanatory text that is shown after the label.
                        REQUIRED (true/false) A boolean value that defines
                        whether the field is required for form submission.
                        DEFAULT is value that is pre-filled in the text field.
  --field-select KEY LABEL HINT REQUIRED DEFAULT KEY_VALUE_PAIRS [KEY_VALUE_PAIRS ...] [KEY LABEL HINT REQUIRED DEFAULT KEY_VALUE_PAIRS [KEY_VALUE_PAIRS ...] ...]
                        A select field. KEY is the meta-data key that stores
                        the field's input. LABEL is the label of the field.
                        HINT is the explanatory text that is shown after the
                        label. REQUIRED (true/false) A boolean value that
                        defines whether the field is required for form
                        submission. DEFAULT is value that is pre-selected.
                        KEY_VALUE_PAIRS is value that is pre-selected.
'''

snapshots['test_block_simple 1'] = '''  - block: ':rocket: Release'
'''

snapshots['test_cli_command 1'] = '''  - command: x

'''

snapshots['test_block_prompt 1'] = '''  - block: ':rocket: Release'
    prompt: Really release?
'''

snapshots['test_block_branches 1'] = '''  - block: ':rocket: Release'
    branches: master release-*
'''

snapshots['test_block_field_text 1'] = '''  - block: ':rocket: Release'
    fields:
      - text: label
        key: key
        hint: hint
        default: default
'''

snapshots['test_block_field_text_required 1'] = '''  - block: ':rocket: Release'
    fields:
      - text: label
        key: key
        hint: hint
        required: true
        default: default
'''

snapshots['test_block_field_text_no_hint 1'] = '''  - block: ':rocket: Release'
    fields:
      - text: label
        key: key
        default: default
'''

snapshots['test_block_field_text_no_default 1'] = '''  - block: ':rocket: Release'
    fields:
      - text: label
        key: key
        hint: hint
'''

snapshots['test_block_field_select 1'] = '''  - block: ':rocket: Release'
    fields:
      - select: label
        key: key
        hint: hint
        default: default
        options:
          - label: Label1
            value: opt1
          - label: Label2
            value: opt2
'''

snapshots['test_block_field_multi_fields 1'] = '''  - block: ':rocket: Release'
    fields:
      - text: label
        key: key
        hint: hint
        default: default
      - select: label
        key: key
        hint: hint
        default: default
        options:
          - label: Label1
            value: opt1
          - label: Label2
            value: opt2
'''



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# pylint: disable=unused-variable
# pylint: disable=missing-docstring

import argparse
import sys
from unittest.mock import patch
import pytest
from bkyml.skeleton import Block, \
                           Comment, \
                           Steps, \
                           Env, \
                           Command, \
                           Plugin, \
                           Wait, \
                           Trigger, \
                           parse_main, \
                           run, \
                           check_positive, \
                           bool_or_string, \
                           plugin_or_key_value_pair

__author__ = "Joscha Feth"
__copyright__ = "Joscha Feth"
__license__ = "mit"


def describe_bkyaml():

    @pytest.fixture
    def run_run(capsys, snapshot, argv):
        with patch.object(sys, 'argv', [''] + argv):
            run()
        captured = capsys.readouterr()
        snapshot.assert_match(captured.out)

    @pytest.fixture
    def args():
        return argparse.Namespace()

    def describe_plugin_or_key_value_pair():
        def test_plugin_or_key_value_pair_plugin():
            assert plugin_or_key_value_pair('org/repo#1.0.0') == 'org/repo#1.0.0'

        def test_plugin_or_key_value_pair_pair():
            assert plugin_or_key_value_pair('a=b=c') == ['a', 'b=c']

    def describe_check_positive():
        def test_check_positive_1():
            assert check_positive(1)

        def test_check_positive_minus_1():
            with pytest.raises(argparse.ArgumentTypeError) as err:
                check_positive(-1)
            assert '-1 is an invalid positive int value' in str(err.value)

    def describe_bool_or_string():
        def test_bool_or_string_true():
            assert bool_or_string('TRUE')

        def test_bool_or_string_false():
            assert bool_or_string('FALSE') is False

        def test_bool_or_string():
            assert bool_or_string('bla') == 'bla'

    def describe_comment():
        def test_comment_missing_str(args, snapshot):
            with pytest.raises(AssertionError):
                Comment.comment(args)

        def test_comment(args, snapshot):
            args.str = ['a', 'b']
            snapshot.assert_match(Comment.comment(args))

        def test_comment_multiline(args, snapshot):
            args.str = ['multiline\ncomments', 'are fun']
            snapshot.assert_match(Comment.comment(args))

    def describe_steps():
        def test_steps(args, snapshot):
            snapshot.assert_match(Steps.steps(args))

    def describe_wait():
        def test_wait(args, snapshot):
            snapshot.assert_match(Wait.wait(args))

        def test_wait_continue_on_failure(args, snapshot):
            args.continue_on_failure = True
            snapshot.assert_match(Wait.wait(args))

    def describe_env():
        def test_env_all(args, snapshot):
            args.var = [['a', 'b'], ['c', 'd']]
            snapshot.assert_match(Env.env(args))

    def describe_block():
        @pytest.fixture
        def generic_block_call(args, snapshot):
            args.label = ':rocket: Release'
            snapshot.assert_match(Block.block(args))

        def test_block_missing(args, snapshot):
            with pytest.raises(AssertionError):
                Block.block(args)

        def test_block_simple(args, snapshot):
            generic_block_call(args, snapshot)

        def test_block_prompt(args, snapshot):
            args.prompt = 'Really release?'
            generic_block_call(args, snapshot)

        def test_block_branches(args, snapshot):
            args.branches = ['master', 'release-*']
            generic_block_call(args, snapshot)

        def test_block_field_text(args, snapshot):
            args.field_text = [['key', 'label', 'hint', 'false', 'default']]
            generic_block_call(args, snapshot)

        def test_block_field_text_no_key(args, snapshot):
            args.field_text = [['', 'label', 'hint', 'false', 'default']]
            with pytest.raises(argparse.ArgumentTypeError):
                generic_block_call(args, snapshot)

        def test_block_field_text_no_label(args, snapshot):
            args.field_text = [['key', '', 'hint', 'false', 'default']]
            with pytest.raises(argparse.ArgumentTypeError):
                generic_block_call(args, snapshot)

        def test_block_field_text_no_hint(args, snapshot):
            args.field_text = [['key', 'label', '', 'false', 'default']]
            generic_block_call(args, snapshot)

        def test_block_field_text_no_default(args, snapshot):
            args.field_text = [['key', 'label', 'hint', 'false', '']]
            generic_block_call(args, snapshot)

        def test_block_field_text_required(args, snapshot):
            args.field_text = [['key', 'label', 'hint', 'true', 'default']]
            generic_block_call(args, snapshot)

        def test_block_field_select(args, snapshot):
            args.field_select = [['key', 'label', 'hint', 'false', 'default', 'opt1=Label1', 'opt2=Label2']]
            generic_block_call(args, snapshot)

        def test_block_field_select_no_key(args, snapshot):
            args.field_select = [['', 'label', 'hint', 'false', 'default', 'opt1=Label1', 'opt2=Label2']]
            with pytest.raises(argparse.ArgumentTypeError):
                generic_block_call(args, snapshot)

        def test_block_field_select_no_label(args, snapshot):
            args.field_select = [['key', '', 'hint', 'false', 'default', 'opt1=Label1', 'opt2=Label2']]
            with pytest.raises(argparse.ArgumentTypeError):
                generic_block_call(args, snapshot)

        def test_block_field_select_no_key_value(args, snapshot):
            args.field_select = [['key', 'label', 'hint', 'false', 'default']]
            with pytest.raises(argparse.ArgumentTypeError):
                generic_block_call(args, snapshot)

        def test_block_field_select_key_value_no_pair(args, snapshot):
            args.field_select = [['key', 'label', 'hint', 'false', 'default', 'opt']]
            with pytest.raises(ValueError):
                generic_block_call(args, snapshot)

        def test_block_field_multi_fields(args, snapshot):
            args.field_text = [['key', 'label', 'hint', 'false', 'default']]
            args.field_select = [['key', 'label', 'hint', 'false', 'default', 'opt1=Label1', 'opt2=Label2']]
            generic_block_call(args, snapshot)

    def describe_trigger():
        @pytest.fixture
        def generic_trigger_call(args, snapshot):
            args.pipeline = 'my-pipeline'
            snapshot.assert_match(Trigger.trigger(args))

        def test_trigger_missing(args, snapshot):
            with pytest.raises(AssertionError):
                Trigger.trigger(args)

        def test_trigger_simple(args, snapshot):
            generic_trigger_call(args, snapshot)

        def test_trigger_label(args, snapshot):
            args.label = ':rocket: Deploy'
            generic_trigger_call(args, snapshot)

        def test_trigger_async(args, snapshot):
            args.async = True
            generic_trigger_call(args, snapshot)

        def test_trigger_branches(args, snapshot):
            args.branches = ['master', 'release-*']
            generic_trigger_call(args, snapshot)

        def test_trigger_build_branch(args, snapshot):
            args.build_branch = 'master'
            generic_trigger_call(args, snapshot)

        def test_trigger_build_commit(args, snapshot):
            args.build_commit = 'c0ffee'
            generic_trigger_call(args, snapshot)

        def test_trigger_build_message(args, snapshot):
            args.build_message = 'Put the lime in the coconut'
            generic_trigger_call(args, snapshot)

        def test_trigger_build_env(args, snapshot):
            args.build_env = [['a', 'b']]
            generic_trigger_call(args, snapshot)

        def test_trigger_build_meta_data(args, snapshot):
            args.build_meta_data = [['a', 'b']]
            generic_trigger_call(args, snapshot)

    def describe_command():
        @pytest.fixture
        def generic_command_call(args, snapshot):
            args.command = [['cmd']]
            snapshot.assert_match(Command.command(args))

        @pytest.fixture
        def assert_command_call_error(capsys, command_args, message):
            with pytest.raises(SystemExit) as sys_exit:
                parse_main(['command', '--command', 'cmd'] + command_args)
            assert '2' in str(sys_exit.value)
            captured = capsys.readouterr()
            assert message in captured.err

        def test_command_1(args, snapshot):
            args.command = [["my-command arg1 'arg 2'"]]
            snapshot.assert_match(Command.command(args))

        def test_command_n(args, snapshot):
            args.command = [['a', 'b'], ['c', 'd']]
            snapshot.assert_match(Command.command(args))

        def test_label(args, snapshot):
            args.label = 'My label'
            generic_command_call(args, snapshot)

        def test_branches(args, snapshot):
            args.branches = ['master', 'release-*']
            generic_command_call(args, snapshot)

        def test_env(args, snapshot):
            args.env = [['a', 'b'], ['c', 'd']]
            generic_command_call(args, snapshot)

        def test_agents(args, snapshot):
            args.agents = [['npm', 'true'], ['mvn', 'true']]
            generic_command_call(args, snapshot)

        def test_artifact_paths_0(args, snapshot):
            args.artifact_paths = []
            generic_command_call(args, snapshot)

        def test_artifact_paths_1(args, snapshot):
            args.artifact_paths = [["logs/**/*;coverage/**/*"]]
            generic_command_call(args, snapshot)

        def test_artifact_paths_n(args, snapshot):
            args.artifact_paths = [["logs/**/*", "coverage/**/*"]]
            generic_command_call(args, snapshot)

        def test_parallelism(args, snapshot):
            args.parallelism = 4
            generic_command_call(args, snapshot)

        def test_parallelism_1(args, snapshot):
            args.parallelism = 1
            generic_command_call(args, snapshot)

        def test_concurrency(args, snapshot):
            args.concurrency = 2
            args.concurrency_group = 'my/group'
            generic_command_call(args, snapshot)

        def test_concurrency_cli(capsys):
            assert_command_call_error(
                capsys,
                ['--concurrency', '1'],
                '--concurrency requires --concurrency-group'
            )

        def test_timeout_in_minutes_minus(args, snapshot):
            args.timeout_in_minutes = -1
            generic_command_call(args, snapshot)

        def test_timeout_in_minutes_0(args, snapshot):
            args.timeout_in_minutes = 0
            generic_command_call(args, snapshot)

        def test_timeout_in_minutes_1(args, snapshot):
            args.timeout_in_minutes = 1
            generic_command_call(args, snapshot)

        def test_skip_bool_true(args, snapshot):
            args.skip = True
            generic_command_call(args, snapshot)

        def test_skip_bool_false(args, snapshot):
            args.skip = False
            generic_command_call(args, snapshot)

        def test_skip_string(args, snapshot):
            args.skip = 'Some reason'
            generic_command_call(args, snapshot)

        def describe_retry():
            def test_retry_unknown(args, snapshot):
                args.retry = 'unknown'
                with pytest.raises(argparse.ArgumentTypeError) as err:
                    generic_command_call(args, snapshot)
                assert 'unknown is an invalid retry value' in str(err.value)

            def describe_manual():
                def test_retry_manual(args, snapshot):
                    args.retry = 'manual'
                    generic_command_call(args, snapshot)

                def test_retry_manual_allowed(args, snapshot):
                    args.retry = 'manual'
                    args.retry_manual_allowed = True
                    generic_command_call(args, snapshot)
                    args.retry_manual_allowed = False
                    generic_command_call(args, snapshot)

                def test_retry_reason(args, snapshot):
                    args.retry = 'manual'
                    args.retry_manual_reason = 'Some reason why'
                    generic_command_call(args, snapshot)

                # pylint: disable=invalid-name
                def test_retry_manual_permit_on_passed(args, snapshot):
                    args.retry = 'manual'
                    args.retry_manual_permit_on_passed = True
                    generic_command_call(args, snapshot)
                    args.retry_manual_permit_on_passed = False
                    generic_command_call(args, snapshot)

                def test_retry_manual_missing_allowed_missing_retry(capsys):
                    assert_command_call_error(
                        capsys,
                        ['--no-retry-manual-allowed'],
                        '--[no-]retry-manual-allowed requires --retry manual'
                    )

                def test_retry_manual_reason_missing_retry(capsys):
                    assert_command_call_error(
                        capsys,
                        ['--retry-manual-reason', 'My reason'],
                        '--retry-manual-reason requires --retry manual'
                    )

                # pylint: disable=invalid-name
                def test_retry_manual_pop_missing_retry(capsys):
                    assert_command_call_error(
                        capsys,
                        ['--retry-manual-permit-on-passed'],
                        '--[no-]retry-manual-permit-on-passed'
                        + ' requires --retry manual'
                    )

                # pylint: disable=invalid-name
                def test_retry_manual_pop_manual_retry(capsys, snapshot):
                    run_run(
                        capsys,
                        snapshot,
                        ['command',
                         '--command', 'cmd',
                         '--retry', 'manual',
                         '--retry-manual-permit-on-passed']
                    )

            def describe_automatic():
                def test_retry_automatic(args, snapshot):
                    args.retry = 'automatic'
                    generic_command_call(args, snapshot)

                def test_retry_automatic_limit(args, snapshot):
                    args.retry = 'automatic'
                    args.retry_automatic_limit = 2
                    generic_command_call(args, snapshot)

                def test_retry_automatic_limit_11(args, snapshot):
                    args.retry = 'automatic'
                    args.retry_automatic_limit = 11
                    generic_command_call(args, snapshot)

                def test_retry_automatic_exit_status_star(args, snapshot):
                    args.retry = 'automatic'
                    args.retry_automatic_exit_status = '*'
                    generic_command_call(args, snapshot)

                def test_retry_automatic_exit_status_number(args, snapshot):
                    args.retry = 'automatic'
                    args.retry_automatic_exit_status = 1
                    generic_command_call(args, snapshot)

                def test_retry_automatic_exit_status_and_limit(args, snapshot):
                    args.retry = 'automatic'
                    args.retry_automatic_limit = 2
                    args.retry_automatic_exit_status = 1
                    generic_command_call(args, snapshot)

                def test_retry_automatic_tuple_0(args, snapshot):
                    args.retry = 'automatic'
                    args.retry_automatic_tuple = []
                    generic_command_call(args, snapshot)

                def test_retry_automatic_tuple_n(args, snapshot):
                    args.retry = 'automatic'
                    args.retry_automatic_tuple = [['*', 2], [1, 3]]
                    generic_command_call(args, snapshot)

                def test_exit_status_string(capsys):
                    assert_command_call_error(
                        capsys,
                        ['--retry', 'automatic',
                         '--retry-automatic-exit-status', 'xxx'],
                        'xxx is an invalid value'
                    )

                def test_retry_exit_status_missing_retry(capsys):
                    assert_command_call_error(
                        capsys,
                        ['--retry-automatic-exit-status', '*'],
                        '--retry-automatic-exit-status'
                        + ' requires --retry automatic'
                    )

                def test_retry_limit_missing_retry(capsys):
                    assert_command_call_error(
                        capsys,
                        ['--retry-automatic-limit', '2'],
                        '--retry-automatic-limit requires --retry automatic'
                    )

                def test_retry_tuple_missing_retry(capsys):
                    assert_command_call_error(
                        capsys,
                        ['--retry-automatic-tuple', '*', '2'],
                        '--retry-automatic-tuple requires --retry automatic'
                    )

                def test_retry_tuple_missing_combine_exit_status(capsys):
                    assert_command_call_error(
                        capsys,
                        ['--retry', 'automatic',
                         '--retry-automatic-tuple', '*', '2',
                         '--retry-automatic-exit-status', '*'],
                        '--retry-automatic-tuple can not be combined with'
                        + ' --retry-automatic-exit-status'
                    )

                def test_retry_tuple_missing_combine_limit(capsys):
                    assert_command_call_error(
                        capsys,
                        ['--retry', 'automatic',
                         '--retry-automatic-tuple', '*', '2',
                         '--retry-automatic-limit', '2'],
                        '--retry-automatic-tuple'
                        + ' can not be combined with '
                        + '--retry-automatic-limit'
                    )

        def describe_command_plugin_attr():

            def test_command_plugin_none(args, snapshot):
                args.plugin = []
                generic_command_call(args, snapshot)

            def test_command_plugin_no_args(args, snapshot):
                args.plugin = [['org/repo#1.0.0']]
                generic_command_call(args, snapshot)

            def test_command_plugin_1(args, snapshot):
                args.plugin = [
                    ['org/repo#1.0.0', ['a', 'b'], ['c', 'd']]
                ]
                generic_command_call(args, snapshot)

            def test_command_plugin_n(args, snapshot):
                args.plugin = [
                    ['org/repo#1.0.0', ['a', 'b'], ['c', 'd']],
                    ['other_org/other_repo', ['e', 'f'], ['g', 'h=i']],
                ]
                generic_command_call(args, snapshot)

    def describe_plugin():

        @pytest.fixture
        def generic_plugin_call(args, snapshot):
            snapshot.assert_match(Plugin.plugin(args))

        def describe_plugin_name_attr():

            def test_plugin_name_attr(args, snapshot):
                args.plugin = [['org/repo#1.0.0']]
                args.name = 'My plugin run'
                generic_plugin_call(args, snapshot)

        def describe_plugin_plugin_attr():

            def test_plugin_plugin_none(args, snapshot):
                args.plugin = []
                generic_plugin_call(args, snapshot)

            def test_plugin_plugin_no_args(args, snapshot):
                args.plugin = [['org/repo#1.0.0']]
                generic_plugin_call(args, snapshot)

            def test_plugin_plugin_1(args, snapshot):
                args.plugin = [
                    ['org/repo#1.0.0', ['a', 'b'], ['c', 'd']]
                ]
                generic_plugin_call(args, snapshot)

            def test_plugin_plugin_n(args, snapshot):
                args.plugin = [
                    ['org/repo#1.0.0', ['a', 'b'], ['c', 'd']],
                    ['other_org/other_repo', ['e', 'f'], ['g', 'h=i']],
                ]
                generic_plugin_call(args, snapshot)

    def describe_parse_main():
        def test_main(snapshot):
            snapshot.assert_match(parse_main(['command', '--command', 'x']))

    def describe_cli():
        def test_cli_command(snapshot, capsys):
            run_run(capsys, snapshot, ['command', '--command', 'x'])

        def test_empty_command(snapshot, capsys):
            with pytest.raises(SystemExit):
                with patch.object(sys, 'argv', ['']):
                    run()
            captured = capsys.readouterr()
            snapshot.assert_match(captured.err)

        def test_help(snapshot, capsys):
            for subcommand in [
                'comment',
                'steps',
                'env',
                'command',
                'plugin',
                'wait',
                'trigger',
                'block'
            ]:
                with pytest.raises(SystemExit):
                    with patch.object(sys, 'argv', ['', subcommand, '--help']):
                        run()
                captured = capsys.readouterr()
                snapshot.assert_match(captured.out)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
'''
    Module to produce valid YAML for Buildkite
    see https://buildkite.com/docs/pipelines/defining-steps
'''
from __future__ import division, print_function, absolute_import

import argparse
import sys
import logging
from ruamel.yaml import YAML as RuamelYaml
from ruamel.yaml.compat import StringIO

from bkyml import __version__

__author__ = "Joscha Feth"
__copyright__ = "Joscha Feth"
__license__ = "mit"

LOGGER = logging.getLogger(__name__)


class MyYAML(RuamelYaml):
    def to_string(self, data):
        stream = StringIO()
        RuamelYaml.dump(self, data, stream)
        return stream.getvalue()


YAML = MyYAML()
YAML.default_flow_style = False

RETRY_MANUAL_ALLOWED_DEFAULT = True
RETRY_MANUAL_PERMIT_ON_PASSED_DEFAULT = False


def plugin_or_key_value_pair(value):
    if '=' in value:
        return value.split('=', 1)
    return value


def check_positive(value):
    ivalue = int(value)
    if ivalue <= 0:
        raise argparse.ArgumentTypeError(
            "%s is an invalid positive int value" % value
        )
    return ivalue


def bool_or_string(value):
    if value.lower() == 'true':
        return True
    if value.lower() == 'false':
        return False
    return value


def int_or_star(value):
    if value == '*':
        return value
    else:
        try:
            return int(value)
        except ValueError:
            raise argparse.ArgumentTypeError("%s is an invalid value" % value)


def ns_hasattr(namespace, attr):
    return hasattr(namespace, attr) and getattr(namespace, attr) is not None


def tuples_to_dict(tuples):
    ret = {}
    for tpl in tuples:
        ret[tpl[0]] = tpl[1]
    return ret


def singlify(a_list):
    if len(a_list) == 1:
        return a_list[0]
    return a_list


def plugins_section(step, namespace):
    if ns_hasattr(namespace, 'plugin') and namespace.plugin:
        plugins = {}

        for plugin in namespace.plugin:
            name, tuples = plugin[0], plugin[1:]
            plugins[name] = None
            if tuples:
                plugins[name] = tuples_to_dict(tuples)
        return plugins
    return None


class Block:

    @staticmethod
    def install(action):
        parser = action.add_parser('block')
        parser.add_argument(
            dest="label",
            help="Label of the block step. Supports emoji.",
            type=str,
            metavar="LABEL")
        parser.add_argument(
            '--prompt',
            help="The instructional message displayed in the dialog box when the unblock step is activated.", # NOQA
            type=str,
            metavar="PROMPT")
        parser.add_argument(
            '--branches',
            help="The branch pattern defining which branches will include this step in their builds.", # NOQA
            type=str,
            nargs='+',
            metavar="BRANCH_PATTERN"
        )
        parser.add_argument(
            '--field-text',
            help="A text field. \
                KEY is the meta-data key that stores the field's input. \
                LABEL is the label of the field. \
                HINT is the explanatory text that is shown after the label.\
                REQUIRED (true/false) A boolean value that defines whether the field is required for form submission. \
                DEFAULT is value that is pre-filled in the text field. \
                ", # NOQA
            type=str,
            nargs=5,
            action='append',
            metavar=('KEY', 'LABEL', 'HINT', 'REQUIRED', 'DEFAULT')
        )
        parser.add_argument(
            '--field-select',
            help="A select field. \
                KEY is the meta-data key that stores the field's input. \
                LABEL is the label of the field. \
                HINT is the explanatory text that is shown after the label.\
                REQUIRED (true/false) A boolean value that defines whether the field is required for form submission. \
                DEFAULT is value that is pre-selected. \
                KEY_VALUE_PAIRS is value that is pre-selected. \
                ", # NOQA
            type=str,
            nargs='+', # TODO: This will produce an incorrect help entry
            action='append',
            metavar='KEY LABEL HINT REQUIRED DEFAULT KEY_VALUE_PAIRS [KEY_VALUE_PAIRS ...]'
        )
        parser.set_defaults(func=Block.block)

    @staticmethod
    def block(namespace):
        assert ns_hasattr(namespace, 'label')
        step = {
            'block': namespace.label
        }

        # prompt
        if ns_hasattr(namespace, 'prompt'):
            step['prompt'] = namespace.prompt

        # branches
        if ns_hasattr(namespace, 'branches'):
            step['branches'] = ' '.join(namespace.branches)

        # Fields
        has_text_fields = ns_hasattr(namespace, 'field_text')
        has_select_fields = ns_hasattr(namespace, 'field_select')

        if has_select_fields or has_text_fields:
            fields = step['fields'] = []
            if has_text_fields:
                for text_field in namespace.field_text:
                    [key, label, hint, required, default] = text_field
                    field = Block.gen_field('text', key, label, hint, required, default)
                    fields.append(field)
            if has_select_fields:
                for select_field in namespace.field_select:
                    if len(select_field) < 6:
                         raise argparse.ArgumentTypeError("Missing required parameters")
                    key, label, hint, required, default = select_field[0], select_field[1], select_field[2], select_field[3], select_field[4]
                    pairs = select_field[5:]
                    field = Block.gen_field('select', key, label, hint, required, default)

                    options = field['options'] = []
                    for pair in pairs:
                        [value, label] = pair.split('=', 1)
                        options.append({
                            'label': label,
                            'value': value,
                        })
                    fields.append(field)

        YAML.indent(sequence=4, offset=2)
        return YAML.to_string([step])

    @staticmethod
    def gen_field(type, key, label, hint, required, default):
        if key is None or key.strip() == '':
            raise argparse.ArgumentTypeError("'%s' is an invalid key" % key)
        if label is None or label.strip() == '':
            raise argparse.ArgumentTypeError("'%s' is an invalid label" % label)
        field = {
            type: label,
            'key': key
        }
        if hint:
            field['hint'] = hint
        if required.lower() == 'true':
            field['required'] = True
        if default:
            field['default'] = default
        return field

class Trigger:

    @staticmethod
    def install(action):
        parser = action.add_parser('trigger')
        parser.add_argument(
            dest="pipeline",
            help="Name of the pipeline to trigger",
            type=str,
            metavar="PIPELINE")
        parser.add_argument(
            '--label',
            dest="label",
            help="The label that will be displayed in the pipeline visualisation in Buildkite. Supports emoji.", # NOQA
            type=str,
            metavar="LABEL")
        parser.add_argument(
            '--async',
            action='store_true',
            help="If given, the step will immediately continue, regardless of the success of the triggered build.", # NOQA
        )
        parser.add_argument(
            '--branches',
            help="The branch pattern defining which branches will include this step in their builds.", # NOQA
            type=str,
            nargs='+',
            metavar="BRANCH_PATTERN"
        )
        parser.add_argument(
            '--build-message',
            help="The message for the build. Supports emoji.",
            type=str,
            metavar="MESSAGE")
        parser.add_argument(
            '--build-commit',
            help="The commit hash for the build",
            type=str,
            metavar="SHA")
        parser.add_argument(
            '--build-branch',
            help="The branch for the build",
            type=str,
            metavar="BRANCH")
        parser.add_argument(
            '--build-env',
            help="A map of environment variables for the triggered build.",
            type=str,
            nargs=2,
            action='append',
            metavar=('KEY', 'VALUE')
        )
        parser.add_argument(
            '--build-meta-data',
            help="A map of meta-data for the triggered build.",
            type=str,
            nargs=2,
            action='append',
            metavar=('KEY', 'VALUE')
        )
        parser.set_defaults(func=Trigger.trigger)

    @staticmethod
    def trigger(namespace):
        assert ns_hasattr(namespace, 'pipeline')
        step = {
            'trigger': namespace.pipeline
        }

        # label
        if ns_hasattr(namespace, 'label'):
            step['label'] = namespace.label

        # async
        if ns_hasattr(namespace, 'async') and namespace.async:
            step['async'] = True

        # branches
        if ns_hasattr(namespace, 'branches'):
            step['branches'] = ' '.join(namespace.branches)

        # build vars
        has_build_branch = ns_hasattr(namespace, 'build_branch')
        has_build_commit = ns_hasattr(namespace, 'build_commit')
        has_build_message = ns_hasattr(namespace, 'build_message')
        has_build_env = ns_hasattr(namespace, 'build_env')
        has_build_meta_data = ns_hasattr(namespace, 'build_meta_data')
        if has_build_branch \
           or has_build_commit \
           or has_build_message \
           or has_build_env \
           or has_build_meta_data:
            build = step['build'] = {}

            if has_build_branch:
                build['branch'] = namespace.build_branch
            if has_build_commit:
                build['commit'] = namespace.build_commit
            if has_build_message:
                build['message'] = namespace.build_message
            if has_build_env:
                build['env'] = tuples_to_dict(namespace.build_env)
            if has_build_meta_data:
                build['meta_data'] = tuples_to_dict(namespace.build_meta_data)

        YAML.indent(sequence=4, offset=2)
        return YAML.to_string([step])


class Plugin:

    @staticmethod
    def install(action):
        parser = action.add_parser('plugin')
        parser.add_argument(
            '--name',
            dest="name",
            help="Name of the plugin step",
            type=str,
            metavar="NAME")
        parser.add_argument(
            '--plugin',
            help="A plugin to run. Optionally key/value pairs for the plugin.",
            dest="plugin",
            nargs='+',
            action='append',
            type=plugin_or_key_value_pair,
            metavar=('PLUGIN', 'KEY_VALUE_PAIR'),
            required=True
        )
        parser.set_defaults(func=Plugin.plugin)

    @staticmethod
    def plugin(namespace):
        step = {}

        if ns_hasattr(namespace, 'name') and namespace.name:
            step['name'] = namespace.name

        plugins = plugins_section(step, namespace)
        if plugins:
            step['plugins'] = plugins
        else:
            return ''

        YAML.indent(sequence=4, offset=2)
        return YAML.to_string([step])


class Comment:

    @staticmethod
    def install(action):
        parser = action.add_parser('comment')
        parser.add_argument(
            dest="str",
            help="Comment",
            type=str,
            nargs='+',
            metavar="COMMENT")
        parser.set_defaults(func=Comment.comment)

    @staticmethod
    def comment(namespace):
        assert ns_hasattr(namespace, 'str')
        lines = "\n# ".join(["\n# ".join(line.splitlines()) for line in namespace.str])
        return f"# {lines}"


class Steps:

    @staticmethod
    def install(action):
        parser = action.add_parser('steps')
        parser.set_defaults(func=Steps.steps)

    @staticmethod
    # pylint: disable=unused-argument
    def steps(namespace):
        return YAML.to_string({'steps': None})


class Env:

    @staticmethod
    def install(action):
        parser = action.add_parser('env')
        parser.add_argument(
            '--var',
            help="A map of environment variables for this pipeline.",
            type=str,
            nargs=2,
            action='append',
            metavar=('KEY', 'VALUE'),
            required=True
        )
        parser.set_defaults(func=Env.env)

    @staticmethod
    def env(namespace):
        return YAML.to_string({
            'env': tuples_to_dict(namespace.var),
        })


class Command:

    # pylint: disable=line-too-long
    @staticmethod
    def install(action):
        parser = action.add_parser('command')
        parser.add_argument(
            '--command',
            help="The shell command/s to run during this step.",
            nargs='+',
            action='append',
            type=str,
            metavar="COMMAND",
            required=True)
        parser.add_argument(
            '--label',
            help="The label that will be displayed in the pipeline visualisation in Buildkite. Supports emoji.", # NOQA
            type=str,
            metavar="LABEL")
        parser.add_argument(
            '--branches',
            help="The branch pattern defining which branches will include this step in their builds.", # NOQA
            type=str,
            nargs='+',
            metavar="BRANCH_PATTERN"
        )
        parser.add_argument(
            '--env',
            help="A map of environment variables for this step.",
            type=str,
            nargs=2,
            action='append',
            metavar=('KEY', 'VALUE')
        )
        parser.add_argument(
            '--agents',
            help="A map of meta-data keys to values to target specific agents for this step.",
            type=str,
            nargs=2,
            action='append',
            metavar=('KEY', 'VALUE')
        )
        parser.add_argument(
            '--artifact-paths',
            help="The glob path or paths where artifacts from this step will be uploaded.",
            nargs='+',
            action='append',
            type=str,
            metavar="GLOB_OR_PATH"
        )
        parser.add_argument(
            '--parallelism',
            help="The number of parallel jobs that will be created based on this step.",
            type=check_positive,
            metavar="POSITIVE_NUMBER"
        )
        parser.add_argument(
            '--concurrency',
            help="The maximum number of jobs created from this step that are allowed to run at the same time. Requires --concurrency-group.", # NOQA
            type=check_positive,
            metavar="POSITIVE_INT"
        )
        parser.add_argument(
            '--concurrency-group',
            help="A unique name for the concurrency group that you are creating with the concurrency attribute.", # NOQA
            type=str,
            metavar="GROUP_NAME"
        )
        parser.add_argument(
            '--timeout-in-minutes',
            help="The number of minutes a job created from this step is allowed to run. If the job does not finish within this limit, it will be automatically cancelled and the build will fail.", # NOQA
            type=int,
            metavar="TIMEOUT"
        )
        parser.add_argument(
            '--skip',
            help="Whether to skip this step or not.",
            type=bool_or_string,
            metavar="BOOL_OR_STRING"
        )
        parser.add_argument(
            '--retry',
            help="The conditions for retrying this step.",
            choices=['automatic', 'manual']
        )
        parser.add_argument(
            '--retry-automatic-exit-status',
            help="The exit status number that will cause this job to retry.",
            type=int_or_star,
            metavar='INT_OR_STAR'
        )
        parser.add_argument(
            '--retry-automatic-limit',
            help="The number of times this job can be retried. The maximum value this can be set to is 10.", # NOQA
            type=check_positive,
            metavar="POSITIVE_INT"
        )
        parser.add_argument(
            '--retry-automatic-tuple',
            help="The exit status number that will cause this job to retry and a limit to go with it.", # NOQA
            type=int_or_star,
            nargs=2,
            action='append',
            metavar=('INT_OR_STAR', 'POSITIVE_INT')
        )
        parser.add_argument(
            '--retry-manual-allowed',
            help="This job can be retried manually.",
            dest="retry_manual_allowed",
            action='store_true'
        )
        parser.add_argument(
            '--no-retry-manual-allowed',
            help="This job can not be retried manually.",
            dest="retry_manual_allowed",
            action='store_false'
        )
        parser.set_defaults(retry_manual_allowed=RETRY_MANUAL_ALLOWED_DEFAULT)
        parser.add_argument(
            '--retry-manual-reason',
            help="A string that will be displayed in a tooltip on the Retry button in Buildkite.",
            type=str,
            metavar='REASON'
        )
        parser.add_argument(
            '--retry-manual-permit-on-passed',
            help="This job can be retried after it has passed.",
            dest="retry_manual_permit_on_passed",
            action='store_true'
        )
        parser.add_argument(
            '--no-retry-manual-permit-on-passed',
            help="This job can not be retried after it has passed.",
            dest="retry_manual_permit_on_passed",
            action='store_false'
        )
        parser.set_defaults(retry_manual_permit_on_passed=RETRY_MANUAL_PERMIT_ON_PASSED_DEFAULT)

        parser.add_argument(
            '--plugin',
            help="A plugin to run with this step. Optionally key/value pairs for the plugin.",
            dest="plugin",
            nargs='+',
            action='append',
            type=plugin_or_key_value_pair,
            metavar=('PLUGIN', 'KEY_VALUE_PAIR')
        )

        parser.set_defaults(func=Command.command)

    @staticmethod
    def assert_post_parse(parsed, parser):
        if ns_hasattr(parsed, 'concurrency') and not ns_hasattr(parsed, 'concurrency_group'):
            parser.error("--concurrency requires --concurrency-group.")

        if not ns_hasattr(parsed, 'retry') or parsed.retry != 'automatic':
            if ns_hasattr(parsed, 'retry_automatic_exit_status'):
                parser.error('--retry-automatic-exit-status requires --retry automatic.')

            if ns_hasattr(parsed, 'retry_automatic_limit'):
                parser.error('--retry-automatic-limit requires --retry automatic.')

            if ns_hasattr(parsed, 'retry_automatic_tuple'):
                parser.error('--retry-automatic-tuple requires --retry automatic.')

        if ns_hasattr(parsed, 'retry_automatic_tuple') and \
           ns_hasattr(parsed, 'retry_automatic_exit_status'):
                parser.error('--retry-automatic-tuple can not be combined with --retry-automatic-exit-status.') # NOQA

        if ns_hasattr(parsed, 'retry_automatic_tuple') and \
           ns_hasattr(parsed, 'retry_automatic_limit'):
                parser.error('--retry-automatic-tuple can not be combined with --retry-automatic-limit.') # NOQA

        if not ns_hasattr(parsed, 'retry') or parsed.retry != 'manual':
            if ns_hasattr(parsed, 'retry_manual_allowed') \
               and parsed.retry_manual_allowed is not RETRY_MANUAL_ALLOWED_DEFAULT:
                parser.error('--[no-]retry-manual-allowed requires --retry manual.')

            if ns_hasattr(parsed, 'retry_manual_permit_on_passed') \
               and parsed.retry_manual_permit_on_passed \
               is not RETRY_MANUAL_PERMIT_ON_PASSED_DEFAULT:
                parser.error('--[no-]retry-manual-permit-on-passed requires --retry manual.')

            if ns_hasattr(parsed, 'retry_manual_reason'):
                parser.error('--retry-manual-reason requires --retry manual.')

    @staticmethod
    def command(namespace):
        step = {}

        # label
        if ns_hasattr(namespace, 'label'):
            step['label'] = namespace.label

        # command (required)
        assert ns_hasattr(namespace, 'command')
        command = sum(namespace.command, [])
        assert command
        step['command'] = singlify(command)

        # branches
        if ns_hasattr(namespace, 'branches'):
            step['branches'] = ' '.join(namespace.branches)

        # env
        if ns_hasattr(namespace, 'env'):
            step['env'] = tuples_to_dict(namespace.env)

        # agents
        if ns_hasattr(namespace, 'agents'):
            step['agents'] = tuples_to_dict(namespace.agents)

        # artifact_paths
        if ns_hasattr(namespace, 'artifact_paths'):
            artifact_paths = sum(namespace.artifact_paths, [])
            if artifact_paths:
                step['artifact_paths'] = singlify(artifact_paths)

        # parallelism
        if ns_hasattr(namespace, 'parallelism'):
            assert check_positive(namespace.parallelism)
            if namespace.parallelism > 1:
                step['parallelism'] = namespace.parallelism

        # concurrency and concurrency_group
        if ns_hasattr(namespace, 'concurrency') and ns_hasattr(namespace, 'concurrency_group'):
            assert check_positive(namespace.concurrency)
            step['concurrency'] = namespace.concurrency
            step['concurrency_group'] = namespace.concurrency_group

        # timeout_in_minutes
        if ns_hasattr(namespace, 'timeout_in_minutes') and namespace.timeout_in_minutes > 0:
            step['timeout_in_minutes'] = namespace.timeout_in_minutes

        # skip
        if ns_hasattr(namespace, 'skip') \
           and (namespace.skip is True or isinstance(namespace.skip, str)):
            step['skip'] = namespace.skip

        # retry
        if ns_hasattr(namespace, 'retry'):
            retry = {}
            retry[namespace.retry] = {}

            if namespace.retry == 'automatic':
                if ns_hasattr(namespace, 'retry_automatic_exit_status'):
                    retry[namespace.retry]['exit_status'] = namespace.retry_automatic_exit_status
                if ns_hasattr(namespace, 'retry_automatic_limit'):
                    retry[namespace.retry]['limit'] = min(10, namespace.retry_automatic_limit)
                if ns_hasattr(namespace, 'retry_automatic_tuple'):
                    if namespace.retry_automatic_tuple:
                        retry[namespace.retry] = []
                        for tpl in namespace.retry_automatic_tuple:
                            retry[namespace.retry].append({
                                'exit_status': int_or_star(tpl[0]),
                                'limit': min(10, check_positive(tpl[1])),
                            })
            elif namespace.retry == 'manual':
                if ns_hasattr(namespace, 'retry_manual_allowed') \
                   and not namespace.retry_manual_allowed:
                    retry[namespace.retry]['allowed'] = namespace.retry_manual_allowed
                if ns_hasattr(namespace, 'retry_manual_reason'):
                    retry[namespace.retry]['reason'] = namespace.retry_manual_reason
                if ns_hasattr(namespace, 'retry_manual_permit_on_passed') \
                   and namespace.retry_manual_permit_on_passed:
                    retry[namespace.retry]['permit_on_passed'] = namespace.retry_manual_permit_on_passed # NOQA
            else:
                raise argparse.ArgumentTypeError("%s is an invalid retry value" % namespace.retry)

            if not retry[namespace.retry]:
                retry[namespace.retry] = True
            step['retry'] = retry

        # plugins
        plugins = plugins_section(step, namespace)
        if plugins:
            step['plugins'] = plugins

        YAML.indent(sequence=4, offset=2)
        return YAML.to_string([step])


class Wait:

    @staticmethod
    def install(action):
        parser = action.add_parser('wait')
        parser.add_argument(
            '--continue-on-failure',
            action='store_true',
            help="Continue even if previous steps failed",
        )
        parser.set_defaults(func=Wait.wait)

    @staticmethod
    def wait(namespace):
        step = 'wait'

        if ns_hasattr(namespace, 'continue_on_failure') and namespace.continue_on_failure:
            step = {
                'wait': None,
                'continue_on_failure': True,
            }

        YAML.indent(sequence=4, offset=2)
        return YAML.to_string([step])


def parse_args(args):
    """Parse command line parameters

    Args:
      args ([str]): command line parameters as list of strings

    Returns:
      :obj:`argparse.Namespace`: command line parameters namespace
    """
    parser = argparse.ArgumentParser(
        description="Generate pipeline YAML for Buildkite")

    subparsers = parser.add_subparsers(title='subcommands',
                                       description='valid subcommands',
                                       help='additional help')

    Comment.install(subparsers)
    Steps.install(subparsers)
    Env.install(subparsers)
    Command.install(subparsers)
    Plugin.install(subparsers)
    Wait.install(subparsers)
    Trigger.install(subparsers)
    Block.install(subparsers)

    parser.add_argument(
        '--version',
        action='version',
        version='bkyml {ver}'.format(ver=__version__))
    parser.add_argument(
        '-v',
        '--verbose',
        dest="loglevel",
        help="set loglevel to INFO",
        action='store_const',
        const=logging.INFO)
    parser.add_argument(
        '-vv',
        '--very-verbose',
        dest="loglevel",
        help="set loglevel to DEBUG",
        action='store_const',
        const=logging.DEBUG)

    if len(sys.argv) == 1:
        parser.print_help(sys.stderr)
        sys.exit(1)

    parsed = parser.parse_args(args)
    Command.assert_post_parse(parsed, parser)
    return parsed


def setup_logging(loglevel):
    """Setup basic logging

    Args:
      loglevel (int): minimum loglevel for emitting messages
    """
    logformat = "[%(asctime)s] %(levelname)s:%(name)s:%(message)s"
    logging.basicConfig(level=loglevel, stream=sys.stdout,
                        format=logformat, datefmt="%Y-%m-%d %H:%M:%S")


def parse_main(args):
    args = parse_args(args)
    setup_logging(args.loglevel)
    return args.func(args)


def main(args):
    """Main entry point allowing external calls

    Args:
      args ([str]): command line parameter list
    """
    LOGGER.debug("Calling function")
    print(parse_main(args))
    LOGGER.info("Script ends here")


def run():
    """Entry point for console_scripts
    """
    main(sys.argv[1:])


if __name__ == "__main__":
    run()



# -*- coding: utf-8 -*-
# pylint: disable=missing-docstring
from pkg_resources import get_distribution, DistributionNotFound

try:
    # Change here if project is renamed and does not equal the package name
    # pylint: disable=invalid-name
    dist_name = __name__
    __version__ = get_distribution(dist_name).version
except DistributionNotFound:  # pragma: no cover
    __version__ = 'unknown'   # pragma: no cover



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
    Setup file for bkyml.

    This file was generated with PyScaffold 3.0.3.
    PyScaffold helps you to put up the scaffold of your new Python project.
    Learn more under: http://pyscaffold.org/
"""

import sys
from setuptools import setup

# Add here console scripts and other entry points in ini-style format
entry_points = """
[console_scripts]
# script_name = bkyml.module:function
# For example:
bkyml = bkyml.skeleton:run
bkyaml = bkyml.skeleton:run
"""


def setup_package():
    needs_sphinx = {'build_sphinx', 'upload_docs'}.intersection(sys.argv)
    sphinx = ['sphinx'] if needs_sphinx else []
    setup(setup_requires=['pyscaffold>=3.0a0,<3.1a0'] + sphinx,
          entry_points=entry_points,
          use_pyscaffold=True)


if __name__ == "__main__":
    setup_package()
